//
// Copyright (c) 2014 CNRS
// Authors: Florent Lamiraux
//
// This file is part of hpp-core
// hpp-core is free software: you can redistribute it
// and/or modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation, either version
// 3 of the License, or (at your option) any later version.
//
// hpp-core is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Lesser Public License for more details.  You should have
// received a copy of the GNU Lesser General Public License along with
// hpp-core  If not, see
// <http://www.gnu.org/licenses/>.

#include <hpp/util/debug.hh>
#include <hpp/model/device.hh>
#include <hpp/core/config-projector.hh>
#include <hpp/core/diffusing-planner.hh>
#include <hpp/core/node.hh>
#include <hpp/core/edge.hh>
#include <hpp/core/path.hh>
#include <hpp/core/path-validation.hh>
#include <hpp/core/problem.hh>
#include <hpp/core/roadmap.hh>
#include <hpp/core/steering-method.hh>
#include "basic-configuration-shooter.hh"

namespace hpp {
  namespace core {
    extern std::string displayConfig (ConfigurationIn_t q);

    DiffusingPlannerPtr_t DiffusingPlanner::createWithRoadmap
    (const Problem& problem, const RoadmapPtr_t& roadmap)
    {
      DiffusingPlanner* ptr = new DiffusingPlanner (problem, roadmap);
      return DiffusingPlannerPtr_t (ptr);
    }

    DiffusingPlannerPtr_t DiffusingPlanner::create (const Problem& problem)
    {
      DiffusingPlanner* ptr = new DiffusingPlanner (problem);
      return DiffusingPlannerPtr_t (ptr);
    }

    DiffusingPlanner::DiffusingPlanner (const Problem& problem):
      PathPlanner (problem),
      configurationShooter_ (new BasicConfigurationShooter (problem.robot ())),
      qProj_ (problem.robot ()->configSize ())
    {
    }

    DiffusingPlanner::DiffusingPlanner (const Problem& problem,
					const RoadmapPtr_t& roadmap) :
      PathPlanner (problem, roadmap),
      configurationShooter_ (new BasicConfigurationShooter (problem.robot ())),
      qProj_ (problem.robot ()->configSize ())
    {
    }

    PathPtr_t DiffusingPlanner::extend (const NodePtr_t& near,
					const ConfigurationPtr_t& target)
    {
      const SteeringMethodPtr_t& sm (problem ().steeringMethod ());
      const ConstraintSetPtr_t& constraints (sm->constraints ());
      if (constraints) {
	ConfigProjectorPtr_t configProjector (constraints->configProjector ());
	if (configProjector) {
	  configProjector->projectOnKernel (*(near->configuration ()), *target,
					    qProj_);
	  if (constraints->apply (qProj_)) {
	    return (*sm) (*(near->configuration ()), qProj_);
	  } else {
	    return PathPtr_t ();
	  }
	}
      }
      return (*sm) (*(near->configuration ()), *target);
    }

    void DiffusingPlanner::oneStep ()
    {
      DevicePtr_t robot (problem ().robot ());
      PathValidationPtr_t pathValidation (problem ().pathValidation ());
      Nodes_t newNodes;
      PathPtr_t validPath, path;
      // Pick a random node
      ConfigurationPtr_t q_rand = configurationShooter_->shoot ();
      hppDout (info, "q_rand = " << displayConfig (*q_rand));
      //
      // First extend each connected component toward q_rand
      //
      bool q_randInRoadmap = false;
      for (ConnectedComponents_t::const_iterator itcc =
	     roadmap ()->connectedComponents ().begin ();
	   itcc != roadmap ()->connectedComponents ().end (); itcc++) {
	// Find nearest node in roadmap
	value_type distance;
	NodePtr_t near = roadmap ()->nearestNode (q_rand, *itcc, distance);
	path = extend (near, q_rand);
	if (path) {
	  bool pathValid = pathValidation->validate (path, false, validPath);
	  // Insert new path to q_near in roadmap
	  value_type t_final = validPath->timeRange ().second;
	  if (t_final != path->timeRange ().first) {
	    ConfigurationPtr_t q_new (new Configuration_t
				      ((*validPath) (t_final)));
	    if (!q_randInRoadmap || !pathValid) {
	      newNodes.push_back (roadmap ()->addNodeAndEdge
				  (near, q_new, validPath));
	    } else {
	      NodePtr_t newNode = roadmap ()->addNode (q_new);
	      roadmap ()->addEdge (near, newNode, validPath);
	      Path::interval_t timeRange = validPath->timeRange ();
	      roadmap ()->addEdge (newNode, near, validPath->extract
				   (Path::interval_t (timeRange.second ,
						      timeRange.first)));
	    }
	    if ((pathValid) && ((*path) (t_final) == *q_rand))
	      q_randInRoadmap = true;
	  }
	}
      }
      //
      // Second try to connect new nodes together
      //
      const SteeringMethodPtr_t& sm (problem ().steeringMethod ());
      if (!q_randInRoadmap) {
	for (Nodes_t::const_iterator itn1 = newNodes.begin ();
	     itn1 != newNodes.end (); ++itn1) {
	  for (Nodes_t::const_iterator itn2 = boost::next (itn1);
	       itn2 != newNodes.end (); ++itn2) {
	    ConfigurationPtr_t q1 ((*itn1)->configuration ());
	    ConfigurationPtr_t q2 ((*itn2)->configuration ());
	    assert (*q1 != *q2);
	    path = (*sm) (*q1, *q2);
	    if (pathValidation->validate (path, false, validPath)) {
	      roadmap ()->addEdge (*itn1, *itn2, path);
	      Path::interval_t timeRange = path->timeRange ();
	      roadmap ()->addEdge (*itn2, *itn1, path->extract
				   (Path::interval_t (timeRange.second,
						      timeRange.first)));
	    }
	  }
	}
      }
    }

    PathVectorPtr_t DiffusingPlanner::finishSolve (const PathVectorPtr_t& path)
    {
      return path;
    }

    void DiffusingPlanner::configurationShooter
    (const ConfigurationShooterPtr_t& shooter)
    {
      configurationShooter_ = shooter;
    }


  } // namespace core
} // namespace hpp
